from pathlib import Path
from tempfile import TemporaryDirectory

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase, override_settings

from content_security_policy import Policy, PolicyList
from content_security_policy.directives import *
from content_security_policy.django.auto_src import AutoHostScriptSrc
from content_security_policy.django.constants import CSP_CONFIG_NAME, CSP_RO_CONFIG_NAME
from content_security_policy.django.management.commands import buildcsp
from content_security_policy.values import *


class BuildCSPTest(TestCase):
    @override_settings(
        **{
            CSP_CONFIG_NAME: [
                AutoHostScriptSrc(
                    *settings.EXTERNAL_SCRIPTS,
                    watch_apps=settings.INSTALLED_APPS,
                ),
            ]
        },
    )
    def test_fail_build(self):
        """
        Command should raise if AutoDirective has unspecified values.
        """
        with self.assertRaises(CommandError), TemporaryDirectory() as tmp_dir:
            csp_path = Path(tmp_dir) / "csp"
            call_command(buildcsp.Command(), path=csp_path)

    @override_settings(
        **{
            CSP_CONFIG_NAME: [
                AutoHostScriptSrc(
                    *settings.EXTERNAL_SCRIPTS,
                    watch_dirs=[Path(__file__).parent / "test_watch_dir"],
                    host="localhost",
                    scheme="http",
                ),
            ]
        },
    )
    def test_watch_dir(self):
        """
        Command should raise if AutoDirective has unspecified values.
        """
        expected_csp = PolicyList(
            Policy(
                ScriptSrc(
                    HostSrc("https://code.jquery.com/jquery-3.5.1.js"),
                    HostSrc("http://localhost/static/bundle.js"),
                    HostSrc("http://localhost/static/index.js"),
                )
            )
        )
        with TemporaryDirectory() as tmp_dir:
            csp_path = Path(tmp_dir) / "csp"
            call_command(buildcsp.Command(), path=csp_path)
            with open(csp_path) as f:
                rendered = f.read()
                self.assertEquals(rendered, str(expected_csp))

    @override_settings(
        **{
            CSP_CONFIG_NAME: [
                AutoHostScriptSrc(
                    *settings.EXTERNAL_SCRIPTS,
                    watch_apps=[
                        "content_security_policy.django",
                    ],
                    host="localhost",
                    scheme="http",
                ),
            ]
        },
    )
    def test_watch_apps(self):
        """
        Command should raise if AutoDirective has unspecified values.
        """
        expected_csp = PolicyList(
            Policy(
                ScriptSrc(
                    HostSrc("https://code.jquery.com/jquery-3.5.1.js"),
                    HostSrc("http://localhost/static/main.js"),
                    HostSrc("http://localhost/static/some-lib.js"),
                )
            )
        )
        with TemporaryDirectory() as tmp_dir:
            csp_path = Path(tmp_dir) / "csp"
            call_command(buildcsp.Command(), path=csp_path)
            with open(csp_path) as f:
                rendered = f.read()
                self.assertEquals(rendered, str(expected_csp))

    @override_settings(
        **{
            CSP_CONFIG_NAME: [
                ScriptSrcAttr(KeywordSource.self),
            ],
        },
    )
    def test_write_csp(self):
        """
        CSP must be written to file.
        """

        # We are not testing render here!
        expected_csp = buildcsp.Command().render()["csp"]

        with TemporaryDirectory() as tmp_dir:
            csp_path = Path(tmp_dir) / "csp"
            call_command(buildcsp.Command(), path=csp_path)
            with open(csp_path, "r") as f:
                file_contents = f.read()
            self.assertEquals(file_contents, str(expected_csp))

    @override_settings(
        **{
            CSP_CONFIG_NAME: None,
            CSP_RO_CONFIG_NAME: [
                ScriptSrc(KeywordSource.self),
            ],
        },
    )
    def test_write_ro_csp(self):
        """
        RO CSP must be written to file.
        """
        # We are not testing render here!
        expected_csp = buildcsp.Command().render()["csp_ro"]

        with TemporaryDirectory() as tmp_dir:
            csp_ro_path = Path(tmp_dir) / "csp_ro"
            call_command(buildcsp.Command(), path_ro=csp_ro_path)
            with open(csp_ro_path, "r") as f:
                file_ro_contents = f.read()
            self.assertEquals(file_ro_contents, str(expected_csp))

    @override_settings(
        **{
            CSP_CONFIG_NAME: [
                StyleSrc(KeywordSource.self),
            ],
            CSP_RO_CONFIG_NAME: [
                ScriptSrcAttr(KeywordSource.self),
            ],
        },
    )
    def test_write_both(self):
        """
        RO CSP must be written to file.
        """
        # We are not testing render here!
        expected_csps = buildcsp.Command().render()

        with TemporaryDirectory() as tmp_dir:
            csp_path = Path(tmp_dir) / "csp"
            csp_ro_path = Path(tmp_dir) / "csp_ro"
            call_command(buildcsp.Command(), path=csp_path, path_ro=csp_ro_path)

            with open(csp_path, "r") as f:
                file_contents = f.read()
            with open(csp_ro_path, "r") as f:
                file_ro_contents = f.read()

            self.assertEquals(file_contents, str(expected_csps["csp"]))
            self.assertEquals(file_ro_contents, str(expected_csps["csp_ro"]))



class CSPError(Exception):
    ...


class NoSuchDirective(CSPError):
    ...


class BadDirectiveValue(CSPError):
    ...


class BadSourceExpression(BadDirectiveValue):
    ...


class BadDirective(CSPError):
    ...


class BadSourceList(BadDirective):
    ...


class BadPolicy(CSPError):
    ...


class ParsingError(CSPError):
    ...



__all__ = [
    # Use for type checking, like in AutoCSPMiddleware
    "AutoSrcDirective",
    # Complete implementations you can actually use in your settings
    "AutoHostScriptSrc",
]

from content_security_policy.directives import ScriptSrc
from content_security_policy.django.auto_src.base import (
    AutoHostSrc,
    AutoSrcDirective,
)


class AutoHostScriptSrc(AutoHostSrc):
    """
    AutoHostSrc for scripts.
    """

    directive = ScriptSrc
    suffix = ".js"



from typing import Sequence, Tuple, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from content_security_policy.directives import Directive
from content_security_policy.django.auto_src import AutoSrcDirective


def get_csp_setting(
    config_name: str,
) -> Tuple[Tuple[Directive | AutoSrcDirective, ...], ...]:
    """
    Get a CSP setting, make sure it's a one- or two-dimensional sequence of
    directives. Raise helpful exceptions.
    :param config_name: Name of setting to parse
    :return: Always a two-dimensional tuple of directives.
    """

    config = getattr(settings, config_name)
    if isinstance(config[0], (tuple, list)):
        if not all(isinstance(el, (tuple, list)) for el in config):
            raise ImproperlyConfigured(
                f"{config_name} needs to be EITHER a sequence (list/tuple) of "
                f"directives OR a sequence of sequences of directives. "
                f"Mixing sequences and directives is ambiguous."
            )
        config = cast(Sequence[Sequence[Directive | AutoSrcDirective]], config)
        return tuple(tuple(el) for el in config)
    else:
        config = cast(Sequence[Directive | AutoSrcDirective], config)
        return (tuple(config),)



from content_security_policy.django.settings import *

HOST = ""


CONTENT_SECURITY_POLICY = [
    AutoHostScriptSrc(
        *EXTERNAL_SCRIPTS,
        host="localhost",
        scheme="http",
        watch_apps=INSTALLED_APPS,
    ),
    ScriptSrcAttr(NoneSrc),
    DefaultSrc(KeywordSource.self),
    StyleSrc(KeywordSource.self),
    FrameAncestors(NoneSrc),
    BaseUri(NoneSrc),
    ObjectSrc(NoneSrc),
    ManifestSrc(NoneSrc),
]

CONTENT_SECURITY_POLICY_REPORT_ONLY = CONTENT_SECURITY_POLICY



from unittest import TestCase

from content_security_policy.parse import (
    directive_from_string,
    policy_from_string,
    policy_list_from_string,
)


class DirectiveParsing(TestCase):
    def test_parse_serialize_directive(self):
        as_string = "sCript-SrC 'self'\t'nonce-FOOBAR'\nhttp://example.com"
        parsed = directive_from_string(as_string)
        self.assertEqual(as_string, str(parsed))

    def test_parse_serialize_policy(self):
        as_string = (
            "default-src https:; script-src\thttps:\x0c'unsafe-inline' \x0c  'unsafe-eval' blob: 'unsafe-inline'; "
            "frame-src https:   data:; style-src\nhttps: 'unsafe-inline'; img-src https: data: blob:; media-src https: "
            "data: blob:; \tfont-src https: data:;cconnect-src https: wss: blob:; child-src https: blob:; object-src "
            "'none'; base-uri https://*.example.com"
        )
        parsed = policy_from_string(as_string)
        self.assertEqual(as_string, str(parsed))

    def test_parse_serialize_policy_list(self):
        as_string = "default-src 'self'; script-src 'nonce-ABCD'; frame-ancestors 'self'; upgrade-insecure-requests;"
        parsed = policy_list_from_string(as_string)
        self.assertEqual(as_string, str(parsed))

    def test_parse_serialize_messy_policy_list(self):
        as_string = "   \n require-trusted-types-for 'script';report-uri /_/somehing/cspreport, script-src 'report-sample' 'nonce-EmTYwW9IZXpvlIOURJMuAQ' 'unsafe-inline';object-src 'none';base-uri 'self';report-uri /_/something/cspreport;worker-src 'self' \t "
        parsed = policy_list_from_string(as_string)
        self.assertEqual(as_string, str(parsed))



from unittest import TestCase
from typing import cast

from content_security_policy import *
from content_security_policy.exceptions import NoSuchDirective
from content_security_policy.parse import policy_from_string


class SimpleExample(TestCase):
    def test_simple_policy(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual(
            "default-src 'self'; frame-ancestors 'self'; object-src 'none'", str(policy)
        )

    def test_add_directive(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        policy += StyleSrc(HostSrc("http://example.com"))
        self.assertEqual(
            "default-src 'self'; frame-ancestors 'self'; object-src 'none'; style-src http://example.com",
            str(policy),
        )

    def test_remove_directive(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        policy -= FrameAncestors
        self.assertEqual(
            "default-src 'self'; object-src 'none'",
            str(policy),
        )

    def test_remove_preserve_whitespace(self):
        policy = policy_from_string(
            "default-src \t 'self'; \n frame-ancestors 'self'; \tobject-src 'none' ;\x0c\t style-src http://example.com"
        )
        policy -= FrameAncestors
        self.assertEqual(
            "default-src \t 'self'; \tobject-src 'none' ;\x0c\t style-src http://example.com",
            str(policy),
        )

    def test_complex_manipulation(self):
        policy = policy_from_string(
            "deFault-src 'self'; Frame-Ancestors\t 'self'; \t object-src 'none'"
        )

        frame_ancestors = policy["frame-ancestors"]
        # alternatively:
        # frame_ancestors = policy.frame_ancestors
        frame_ancestors += HostSrc("https://example.com")

        policy -= FrameAncestors
        policy += frame_ancestors

        self.assertEqual(
            "deFault-src 'self'; \t object-src 'none'; Frame-Ancestors\t 'self' https://example.com",
            str(policy),
        )

    def test_getitem_int(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual("default-src 'self'", str(policy[0]))

    def test_getitem_str_kebab(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual("frame-ancestors 'self'", str(policy["frame-ancestors"]))

    def test_getitem_str_pascal(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual("frame-ancestors 'self'", str(policy["FrameAncestors"]))

    def test_getitem_type(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual("object-src 'none'", str(policy[ObjectSrc]))

    def test_getitem_type_error(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        with self.assertRaises(TypeError):
            key = cast(str, ValueItem)
            _ = policy[key]

    def test_getitem_no_such_directive(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        with self.assertRaises(NoSuchDirective):
            _ = policy["does-not-exist"]

    def test_getitem_index_error(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        with self.assertRaises(IndexError):
            _ = policy["style-src"]

    def test_getattr_snake(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual("object-src 'none'", str(policy.object_src))

    def test_getattr_Pascal(self):
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        self.assertEqual("object-src 'none'", str(policy.ObjectSrc))

    def test_getitem_attr_error_not_found(self):
        """
        Attribute error should be raised if the policy doesn't have such a directive.
        """
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        with self.assertRaises(AttributeError):
            _ = policy.style_src

    def test_getitem_attr_error_invalid(self):
        """
        Attribute error should be raised instead of NoSuchDirective.
        """
        policy = Policy(
            DefaultSrc(KeywordSource.self), FrameAncestors(SelfSrc), ObjectSrc(NoneSrc)
        )
        with self.assertRaises(AttributeError):
            _ = policy.nonexisting_directive



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test import Client, SimpleTestCase, override_settings

from content_security_policy.constants import CSP_HEADER, CSP_RO_HEADER
from content_security_policy.directives import *
from content_security_policy.django.auto_src import AutoHostScriptSrc
from content_security_policy.django.constants import CSP_CONFIG_NAME, CSP_RO_CONFIG_NAME
from content_security_policy.django.middleware import AutoCSPMiddleware
from content_security_policy.values import *

TEST_CSP_SETTING = [
    AutoHostScriptSrc(
        *settings.EXTERNAL_SCRIPTS,
        watch_apps=settings.INSTALLED_APPS,
    ),
    ScriptSrcAttr(NoneSrc),
    DefaultSrc(KeywordSource.self),
    StyleSrc(KeywordSource.self),
    FrameAncestors(NoneSrc),
    BaseUri(NoneSrc),
    ObjectSrc(NoneSrc),
    ManifestSrc(NoneSrc),
]

TEST_CSP_RO_SETTING = [
    ScriptSrcAttr(KeywordSource.self),
    DefaultSrc(KeywordSource.self),
    StyleSrc(KeywordSource.self),
]


class AutoCSPMiddlewareTests(SimpleTestCase):
    def test_no_debug(self):
        """
        Tests set DEBUG=False by default, the middleware must raise.
        """
        with self.assertRaises(ImproperlyConfigured):
            self.client.get("/")

    @override_settings(DEBUG=True)
    def test_no_host_header(self):
        """
        The middleware uses the host header to construct source expressions.
        Return 400 if it is not set.
        """
        response = self.client.get("/")
        self.assertEquals(response.status_code, 400)

    @override_settings(
        DEBUG=True,
        **{
            CSP_CONFIG_NAME: TEST_CSP_SETTING,
        },
    )
    def test_csp_inject(self):
        """
        CSP header needs to be injected.
        This test is ONLY for the injection mechanism, not for rendering.
        """
        expected_csp = AutoCSPMiddleware.render(
            [TEST_CSP_SETTING],
            scheme="http",
            host="testserver",
        )
        client = Client(HTTP_HOST="testserver")
        response = client.get("/")
        self.assertEquals(response.headers[CSP_HEADER], str(expected_csp))

    @override_settings(
        DEBUG=True,
        **{
            CSP_RO_CONFIG_NAME: TEST_CSP_RO_SETTING,
        },
    )
    def test_csp_ro_inject(self):
        """
        CSP RO header needs to be injected.
        This test is ONLY for the injection mechanism, not for rendering.
        """
        expected_csp_ro = AutoCSPMiddleware.render(
            [TEST_CSP_RO_SETTING],
            scheme="http",
            host="testserver",
        )
        client = Client(HTTP_HOST="testserver")
        response = client.get("/")
        self.assertEquals(response.headers[CSP_RO_HEADER], str(expected_csp_ro))

    @override_settings(
        DEBUG=True,
        **{
            CSP_CONFIG_NAME: TEST_CSP_SETTING,
            CSP_RO_CONFIG_NAME: TEST_CSP_RO_SETTING,
        },
    )
    def test_csp_both_inject(self):
        """
        Both CSP headers needs to be injected.
        This test is ONLY for the injection mechanism, not for rendering.
        """
        expected_csp = AutoCSPMiddleware.render(
            [TEST_CSP_SETTING],
            scheme="http",
            host="testserver",
        )
        expected_csp_ro = AutoCSPMiddleware.render(
            [TEST_CSP_RO_SETTING],
            scheme="http",
            host="testserver",
        )
        client = Client(HTTP_HOST="testserver")
        response = client.get("/")
        self.assertEquals(response.headers[CSP_HEADER], str(expected_csp))
        self.assertEquals(response.headers[CSP_RO_HEADER], str(expected_csp_ro))



from abc import ABCMeta

import string
from typing import Iterable


def kebab_to_pascal(text: str) -> str:
    return string.capwords(text, "-").replace("-", "")


def kebab_to_snake(text: str) -> str:
    return "_".join(text.split("-"))


class StrOnClassMeta(ABCMeta):
    _value: str

    def __str__(cls):
        """
        Calling str() on the CLASS will return  _value.
        """
        return cls._value


class AutoInstanceMixin:
    _auto_instance_prop: Iterable[str] = tuple()

    def __init_subclass__(cls, **kwargs):
        for name in cls._auto_instance_prop:
            prop_name = name.strip("'").replace("-", "_")

            @classmethod
            @property
            def factory(cls, sneak_me=name):
                return cls(sneak_me)

            setattr(cls, prop_name, factory)

        delattr(cls, "_auto_instance_prop")
        super().__init_subclass__(**kwargs)



from unittest import TestCase

from content_security_policy.values import KeywordSource, NoneSrc
from content_security_policy.constants import KEYWORD_SOURCES, NONE
from content_security_policy.utils import kebab_to_snake


class KeywordSourceAttributes(TestCase):
    def test_keyword_source_attributes(self):
        for kw_src in KEYWORD_SOURCES:
            with self.subTest(kw_src):
                attr_name = kebab_to_snake(kw_src).strip("'")
                if not hasattr(KeywordSource, attr_name):
                    self.fail(f"{KeywordSource.__name__} has no attribute {attr_name}")
                elif not isinstance(getattr(KeywordSource, attr_name), KeywordSource):
                    self.fail(
                        f"{KeywordSource.__name__}.{attr_name} is not an instance of {KeywordSource}"
                    )
                else:
                    self.assertEqual(kw_src, str(getattr(KeywordSource, attr_name)))


class NoneSourceStr(TestCase):
    def test_instance_str(self):
        self.assertEqual(NONE, str(NoneSrc()))

    def test_class_str(self):
        self.assertEqual(NONE, str(NoneSrc))

    def test_value_constructor(self):
        as_str = "'NOnE'"
        instance = NoneSrc(_value=as_str)
        self.assertEqual(as_str, str(instance))



__all__ = [
    "TOKEN",
    "BASE64_VALUE",
    "SCHEME",
    "NONCE_SOURCE",
    "HASH_SOURCE",
    "SCHEME_SOURCE",
    "HOST_SOURCE",
    "NONE_SOURCE",
    "SELF_SOURCE",
    "URI_REFERENCE",
    "KEYWORD_SOURCE",
    "WEBRTC_VALUE",
    "SANDBOX_VALUE",
    "ASCII_WHITESPACE",
    "NOT_SEPARATOR",
    "VALUE_ITEM_SEPARATOR",
    "DIRECTIVE_SEPARATOR",
    "POLICY_SEPARATOR",
    "WHITESPACE_HEAD",
]
# These expressions will be compiled with re.IGNORECASE
__case_insensitive__ = {
    "SANDBOX_VALUE",
    "NONCE_SOURCE",
    "HASH-SOURCE",
    "NONE_SOURCE",
    "SELF_SOURCE",
}

import re
from typing import cast

from content_security_policy.constants import (
    HASH_ALGORITHMS,
    KEYWORD_SOURCES,
    NONE,
    SANDBOX_VALUES,
    SELF,
    WEBRTC_VALUES,
)

# https://tools.ietf.org/html/rfc5234#appendix-B.1
ALPHA = cast(re.Pattern, r"[A-Za-z]")
DIGIT = cast(re.Pattern, r"[0-9]")

# https://w3c.github.io/webappsec-csp/#grammardef-base64-value
BASE64_VALUE = cast(re.Pattern, rf"({ALPHA}|{DIGIT}|[+\/\-_]){{2, 0}}={{0, 2}}")
NONCE_SOURCE = cast(re.Pattern, f"'nonce-{BASE64_VALUE}'")
HASH_SOURCE = cast(
    re.Pattern, f"'({'|'.join(alg for alg in HASH_ALGORITHMS)})-{BASE64_VALUE}'"
)

# https://datatracker.ietf.org/doc/html/rfc9110#section-5.6.2
TOKEN_CHAR = f"[!#$%&'*+\-.^_`|~]|{ALPHA}|{DIGIT}"
TOKEN = cast(re.Pattern, f"({TOKEN_CHAR})+")

# https://datatracker.ietf.org/doc/html/rfc3986#appendix-A
UNRESERVED = f"({ALPHA}|{DIGIT}|[-._~])"
HEXDIG = "[0-9a-fA-F]"
PCT_ENCODED = f"%{HEXDIG}{HEXDIG}"
# Deviating from rfc3986 here, since CSP explicitly excludes ";" and "," from ALL directive values:
# https://w3c.github.io/webappsec-csp/#framework-directives
SUB_DELIMS = "[!$&'()*+=]"
PCHAR = f"({UNRESERVED}|{PCT_ENCODED}|{SUB_DELIMS}|@|:)"
SEGMENT = f"{PCHAR}*"
SEGMENT_NZ = f"{PCHAR}+"  # Non-Zero
PATH_ABSOLUTE = f"/({SEGMENT_NZ}(/{SEGMENT})*)?"
SCHEME = cast(re.Pattern, rf"{ALPHA}({ALPHA}|{DIGIT}|[+-.])*")
DEC_OCTET = f"({DIGIT})|([1-9]{DIGIT})|(1{DIGIT}{{2}})|(2[0-4]{DIGIT})|(25[0-5])"
IP_V4_ADDRESS = f"{DEC_OCTET}.{DEC_OCTET}.{DEC_OCTET}.{DEC_OCTET}"
H16 = f"{HEXDIG}{{1,4}}"
LS_32 = f"({H16}:{H16})|({IP_V4_ADDRESS})"
IP_V6_ADDRESS = (
    f"({H16}:){{6}}{LS_32}"
    f"::({H16}:){{5}}{LS_32}"
    f"({H16})?::({H16}:){{4}}{LS_32}"
    f"(({H16}){{0,1}}{H16})?::({H16}:){{3}}{LS_32}"
    f"(({H16}){{0,2}}{H16})?::({H16}:){{2}}{LS_32}"
    f"(({H16}){{0,3}}{H16})?::{H16}:{LS_32}"
    f"(({H16}){{0,4}}{H16})?::{LS_32}"
    f"(({H16}){{0,5}}{H16})?::{H16}"
    f"(({H16}){{0,6}}{H16})?::"
)
# IP_V_FUTURE = # By the time  this is relevant, I will be dead
IP_LITERAL = rf"\[{IP_V6_ADDRESS}\]"  # Fixme when I am dead: f"[({IP_V6_ADDRESS})|({IP_V_FUTURE})]"
REG_NAME = f"({UNRESERVED}|{PCT_ENCODED}|{SUB_DELIMS})*"
HOST = f"({IP_LITERAL})|({IP_V4_ADDRESS})|({REG_NAME})"
PORT = f"({DIGIT})*"
USERINFO = f"({UNRESERVED}|{PCT_ENCODED}|{SUB_DELIMS}|:)*"
AUTHORITY = f"({USERINFO})?({HOST})(:{PORT})?"
FRAGMENT = f"({PCHAR}|[/?])*"
QUERY = f"({PCHAR}|[/?])*"
PATH_AB_EMPTY = f"(/{SEGMENT})*"
PATH_ROOTLESS = f"{SEGMENT_NZ}(/{SEGMENT})*"
SEGMENT_NZ_NC = f"({UNRESERVED}|{PCT_ENCODED}|{SUB_DELIMS}|@)+"
PATH_NOSCHEME = f"{SEGMENT_NZ_NC}(/{SEGMENT})*"
HIER_PART = f"((//{AUTHORITY}{PATH_AB_EMPTY})|{PATH_ABSOLUTE}|{PATH_ROOTLESS})?"
URI = rf"{SCHEME}:{HIER_PART}(\?{QUERY})?(#{FRAGMENT})"
RELATIVE_PART = f"((//{AUTHORITY}{PATH_AB_EMPTY})|({PATH_ABSOLUTE})|({PATH_NOSCHEME}))?"
RELATIVE_REF = rf"({RELATIVE_PART})(\?{QUERY})?(#{FRAGMENT})?"
URI_REFERENCE = cast(re.Pattern, f"({URI})|({RELATIVE_REF})")

# https://w3c.github.io/webappsec-csp/#grammardef-scheme-source
SCHEME_SOURCE = cast(re.Pattern, f"{SCHEME}:")

# https://w3c.github.io/webappsec-csp/#grammardef-host-source
SCHEME_PART = SCHEME
HOST_CHAR = f"({ALPHA}|{DIGIT}|-)"
HOST_PART = rf"(\*|(\*\.)?{HOST_CHAR}+(\.{HOST_CHAR}+)*)"
PORT_PART = rf"(\*|{DIGIT}+)"
PATH_PART = PATH_ABSOLUTE
HOST_SOURCE = cast(
    re.Pattern, f"({SCHEME_PART}://)?{HOST_PART}(:{PORT_PART})?({PATH_PART})?"
)
# https://w3c.github.io/webappsec-csp/#grammardef-keyword-source
KEYWORD_SOURCE = cast(re.Pattern, "|".join(KEYWORD_SOURCES))
NONE_SOURCE = cast(re.Pattern, NONE)

# https://w3c.github.io/webappsec-csp/#grammardef-ancestor-source-list
SELF_SOURCE = cast(re.Pattern, SELF)

# https://w3c.github.io/webappsec-csp/#directive-webrtc
WEBRTC_VALUE = cast(re.Pattern, "|".join(WEBRTC_VALUES))

# https://w3c.github.io/webappsec-csp/#directive-sandbox
SANDBOX_VALUE = cast(re.Pattern, "|".join(SANDBOX_VALUES))

# https://w3c.github.io/webappsec-csp/#grammardef-optional-ascii-whitespace
WHITESPACE_CHARS = "\t\n\x0c\r "
ASCII_WHITESPACE = cast(re.Pattern, f"[{WHITESPACE_CHARS}]")

VALUE_ITEM_SEPARATOR = cast(re.Pattern, f"{ASCII_WHITESPACE}+")
DIRECTIVE_SEPARATOR = cast(re.Pattern, f"{ASCII_WHITESPACE}*;{ASCII_WHITESPACE}*")

POLICY_SEPARATOR = cast(re.Pattern, f"{ASCII_WHITESPACE}*,{ASCII_WHITESPACE}*")
# Used for unrecognized directive names / hash items
NOT_SEPARATOR = cast(re.Pattern, f"[^{WHITESPACE_CHARS};,]*")
WHITESPACE_HEAD = re.compile(f"^{ASCII_WHITESPACE}*", flags=re.MULTILINE)

# workaround for the "want to reuse patters but also want to precompile them"-problem
for name in __all__:
    pat = locals()[name]
    if isinstance(pat, re.Pattern):
        continue
    if name in __case_insensitive__:
        locals()[name] = re.compile(pat, flags=re.IGNORECASE)
    else:
        locals()[name] = re.compile(pat)


# Convenience for debug
if __name__ == "__main__":
    for name in __all__:
        print(name)
        print(locals()[name].pattern)



# https://w3c.github.io/webappsec-csp/#csp-directives
FETCH_DIRECTIVE_NAMES = (
    "child-src",
    "connect-src",
    "default-src",
    "font-src",
    "frame-src",
    "img-src",
    "manifest-src",
    "media-src",
    "object-src",
    "script-src",
    "script-src-elem",
    "script-src-attr",
    "style-src",
    "style-src-elem",
    "style-src-attr",
)
DOCUMENT_DIRECTIVE_NAMES = ("base-uri", "sandbox")
NAVIGATION_DIRECTIVE_NAMES = ("form-action", "frame-ancestors")
REPORTING_DIRECTIVE_NAMES = ("report-uri", "report-to")
OTHER_DIRECTIVE_NAMES = ("webrtc", "worker-src")

DIRECTIVE_NAMES = (
    FETCH_DIRECTIVE_NAMES
    + DOCUMENT_DIRECTIVE_NAMES
    + NAVIGATION_DIRECTIVE_NAMES
    + REPORTING_DIRECTIVE_NAMES
    + OTHER_DIRECTIVE_NAMES
)

# All these directives have a serialized-source-list as hash
SOURCE_LIST_DIRECTIVES = FETCH_DIRECTIVE_NAMES + (
    "worker-src",
    "form-action",
    "base-uri",
)

# https://w3c.github.io/webappsec-csp/#grammardef-hash-algorithm
HASH_ALGORITHMS = (
    "sha256",
    "sha384",
    "sha512",
)

# https://w3c.github.io/webappsec-csp/#grammardef-keyword-source
KEYWORD_SOURCES = (
    "'self'",
    "'unsafe-inline'",
    "'unsafe-eval'",
    "'strict-dynamic'",
    "'unsafe-hashes'",
    "'report-sample'",
    "'unsafe-allow-redirects'",
    "'wasm-unsafe-eval'",
)

# https://w3c.github.io/webappsec-csp/#directive-webrtc
WEBRTC_VALUES = ["'allow'", "'block'"]

# https://w3c.github.io/webappsec-csp/#grammardef-nonce-source
NONCE_PREFIX = "nonce-"

NONE = "'none'"

SELF = "'self'"

DEFAULT_VALUE_SEPARATOR = " "
DEFAULT_DIRECTIVE_SEPARATOR = "; "
DEFAULT_POLICY_SEPARATOR = ", "


# https://html.spec.whatwg.org/multipage/iframe-embed-object.html#the-iframe-elemet
SANDBOX_VALUES = (
    "allow-downloads",
    "allow-forms",
    "allow-modals",
    "allow-orientation-lock",
    "allow-pointer-lock",
    "allow-popups",
    "allow-popups-to-escape-sandbox",
    "allow-presentation",
    "allow-same-origin",
    "allow-scripts",
    "allow-top-navigation",
    "allow-top-navigation-by-user-activation",
    "allow-top-navigation-to-custom-protocols",
)

CSP_HEADER = "Content-Security-Policy"
CSP_RO_HEADER = "Content-Security-Policy-Report-Only"



from abc import ABCMeta, abstractmethod
from functools import cache
from pathlib import Path
from typing import *

from django.apps import apps
from django.templatetags.static import static
from watchdog.events import (
    FileCreatedEvent,
    FileDeletedEvent,
    FileModifiedEvent,
    FileMovedEvent,
    FileSystemEvent,
    FileSystemEventHandler,
)

from content_security_policy.directives import Directive
from content_security_policy.django.exceptions import ValuesMissing
from content_security_policy.values import HostSrc, SourceExpression

SRC_HASH_FUN = "sha384"


def fill_render_args(*argnames: str, optional: List[str] = None):
    """
    Decorator that overrides kwargs for render with equally named instance attributes.
    """

    def decorator(render_func: Callable) -> Callable:
        def wrapped_render(self: AutoSrcDirective, *args, **kwargs):
            missing = []
            for arg in argnames:
                on_self = getattr(self, arg)
                in_args = kwargs.get(arg)
                if on_self is None and in_args is None:
                    missing.append(arg)
                kwargs[arg] = on_self or in_args

            if optional is not None:
                for arg in optional:
                    kwargs[arg] = kwargs.get(arg) or getattr(self, arg)

            if missing:
                raise ValuesMissing(
                    f"Attempted to render a {self.name}, but some value(s) "
                    "where neither found on the policies attributes nor in the "
                    f"render arguments: {', '.join(missing)}",
                    names=missing,
                )

            return render_func(self, *args, **kwargs)

        return wrapped_render

    return decorator


def init_before_first_render(render_func):
    """
    Calls self.init_files before the first execution of render_func.
    Use it to call init_files before the first run of render.
    """

    def new_render(self, **kwargs):
        try:
            self.files_initialized
        except AttributeError:
            self.init_files()
            self.files_initialized = True
        return render_func(self, **kwargs)

    return new_render


DirectiveType = TypeVar("DirectiveType", bound=Directive)
# This is not bound to ValueItem, because you may have to defer creating an actual
# ValueItem until render
DynamicValueType = TypeVar("DynamicValueType")


class AutoSrcDirective(
    FileSystemEventHandler, Generic[DirectiveType, DynamicValueType], metaclass=ABCMeta
):
    """
    Automatically generates a -src directive from a list of source static_values and
    files in watch_dirs. Subclasses need to provide:
    - a directive class
    - a suffix (file extension) that will be "watched"
    - a method to compute a directive value that allow-lists a file in watch_dirs
      whenever it changes: compute_value_item
    """

    @property
    @abstractmethod
    def suffix(self) -> str:
        """
        Suffix (file extension) of the type of file that the directive watches for.
        """
        ...

    @property
    @abstractmethod
    def directive(self) -> Type[Directive]:
        """
        Directive class to render.
        """
        ...

    @abstractmethod
    def compute_value_item(self, path: Path) -> DynamicValueType:
        """
        Compute a directive value item to allow-list a local file.
        """
        ...

    def __init__(
        self,
        *static_values: Tuple[SourceExpression],
        use_self_keyword: bool = False,
        watch_dirs: List[Path | str] | None = None,
        watch_apps: List[str] | None = None,
    ):
        self.static_values = static_values
        self.use_self_keyword = use_self_keyword
        self._watch_dirs = [Path(d) for d in watch_dirs] if watch_dirs else []
        self._watch_apps = watch_apps or []

        # Computing values for files is deferred until the first render,
        # because you might need django to be fully started up to compute anyting
        # E.g.: AutoHostSrc calls django.templatetags.static.static, which can only be
        # called once django is fully up. See self.init_files.
        self.files: Dict[Path, DynamicValueType] = {}

    @property
    def name(self):
        """
        Return _name of directive that this auto directive will render to.

        "_name" is used instead of "name because:
        self.directive.name is a property object - not a meaningful value because
        directive is a directive TYPE, not an instance.
        Directive.name can not become a classproperty because that would disable the
        ability to override directive names.
        """
        return self.directive._name

    @property
    @cache
    def watch_dirs(self):
        app_static_dirs = []
        app_paths = {conf.name: conf.path for conf in apps.app_configs.values()}
        for app in self._watch_apps:
            expected_static = Path(app_paths[app]) / "static"
            if expected_static.is_dir():
                app_static_dirs.append(expected_static)
        absolute_watch_dirs = [dir.absolute() for dir in self._watch_dirs]

        return list(set(absolute_watch_dirs + app_static_dirs))

    def on_any_event(self, event: FileSystemEvent):
        """
        Handle all FileSystemEvents from Watchdog. For newly created, modified or moved
        files, call compute_value. Remove directive static_values for deleted files.
        """
        resource = Path(event.src_path)

        if resource.is_dir() or resource.suffix != self.suffix:
            return

        if isinstance(event, FileMovedEvent):
            # Remove hash of old file
            self.files.pop(resource.absolute(), None)
            # Then we just treat it like creation / modification of the "destination"
            resource = Path(event.dest_path)

        if isinstance(event, (FileModifiedEvent, FileCreatedEvent, FileMovedEvent)):
            self.files[resource.absolute()] = self.compute_value_item(resource)
        if isinstance(event, FileDeletedEvent):
            self.files.pop(resource.absolute(), None)

    def init_files(self):
        """
        Make sure all local files have a value in self.files.
        """
        initial_paths: Set[Path] = {
            file
            for watch_dir in self.watch_dirs
            for file in watch_dir.rglob(f"*{self.suffix}")
        }
        for p in initial_paths:
            if p not in self.files:
                self.files[p] = self.compute_value_item(p)

    @init_before_first_render
    def render(self, **kwargs) -> DirectiveType:
        # Sorting to avoid non-deterministic tests
        values = sorted(self.files.values())
        return self.directive(*self.static_values, values)


class AutoHostSrc(AutoSrcDirective[DirectiveType, str], metaclass=ABCMeta):
    """
    Computes "source-expressions" for a -src CSP. For local files the URL is
    constructed. For external sources their src-attribute is used. Host and scheme for
    local files are added when rendered.
    """

    def __init__(
        self,
        *static_values: Tuple[SourceExpression],
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(*static_values, **kwargs)
        self.scheme = scheme
        self.host = host
        self.port = port

    def compute_value_item(self, path: Path) -> str:
        """
        Create relative url for -src directive, the complete URL is computed in render.
        """
        for stat_dir in self.watch_dirs:
            if path.is_relative_to(stat_dir):
                return static(str(path.relative_to(stat_dir)))

        raise ValueError(
            f"Could not find {path} in any watch dir. Probably a watchdog observer is"
            " misconfigured."
        )

    @init_before_first_render
    @fill_render_args("scheme", "host", optional=["port"])
    def render(
        self,
        *,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        **kwargs,
    ) -> DirectiveType:
        if scheme is None or host is None:
            missing = {"scheme": scheme, "host": host}
            missing = [k for k, v in missing.items() if v is None]
            raise ValuesMissing(
                f"Attempted to render urls for {self.name}, but some "
                f"required value(s) where neither found on the directives attributes "
                f"nor in the render arguments: {', '.join(missing)}",
                names=missing,
            )

        origin = f"{scheme}://{host}"

        if port is not None:
            origin = f"{origin}:{port}"

        # Sorting to avoid non-deterministic tests
        paths = sorted(self.files.values())
        dynamic = [HostSrc(f"{origin}{path}") for path in paths]

        return self.directive(*self.static_values, *dynamic)


# TODO: Revisit
#  I only discovered after implementing AutoHashSrc that Firefox only supports cksums
#  for inline elements and NOT for anything loaded via URLs. The use-case for this
#  AutoDirective would thus be limited to chrome and break firefox without adding
#  more source expressions ('self' or HostSrces) to allow the tracked files.
#  That would in turn defeat any security benefit gained from using cksums.
#  Different CSPs for different browsers is something I do not want to advocate since
#  it contradicts the core idea of this project:
#  Nail your CSP down in one place and deploy it as statically as possible!


# from content_security_policy.django.utils import get_file_hash
# from content_security_policy.values import HashSrc
# class AutoHashSrc(AutoSrcDirective[DirectiveType, HashSrc], metaclass=ABCMeta):
#     """
#     Computes "source-hashes" for a -src CSP. Local files are hashed whenever they
#     change.
#     """
#
#     def compute_value_item(self, path: Path) -> HashSrc:
#         """
#         Compute hash of file and create value for -src directive.
#         """
#
#         return HashSrc(
#             f"'{SRC_HASH_FUN}-" f"{get_file_hash(path, hash_name=SRC_HASH_FUN)}'"
#         )



__all__ = ["get_file_hash"]

import hashlib
import mmap
from pathlib import Path
from sys import platform

CHUNK_SIZE = 1024 * 64  # 64 kb


# Newer versions of py hashlib have dedicated function for hashing files
def get_file_hash(path: Path, hash_name: str = "sha256") -> str:
    hash_factory = getattr(hashlib, hash_name)
    hash_obj = hash_factory()

    if path.stat().st_size != 0:  # Can not mmap empty files
        with open(path, "rb") as f:
            if platform == "linux":
                # Mmap significantly speeds up hashing while keeping memory footprint small
                map = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
                hash_obj.update(map)
            else:  # TODO: mmap for windows and OSX
                while data := f.read(CHUNK_SIZE):
                    hash_obj.update(data)

    return hash_obj.hexdigest()



from __future__ import annotations

__all__ = [
    "ValueItem",
    "ValueItemType",
    "NonceSrc",
    "HashSrc",
    "SchemeSrc",
    "HostSrc",
    "KeywordSource",
    "SourceExpression",
    "SourceList",
    "WebrtcValue",
    "NoneSrc",
    "NoneSrcType",
    "SelfSrc",
    "SelfSrcType",
    "AncestorSource",
    "AncestorSourceList",
    "SandboxToken",
    "SandboxValue",
    "ReportToValue",
    "UriReference",
    "ReportUriValue",
    "UnrecognizedValueItem",
]

from abc import ABC
from typing import Literal, Optional, Tuple, Type, Union, cast

from content_security_policy.base_classes import (
    ClassAsValue,
    SelfType,
    ValueItem,
    ValueItemType,
)
from content_security_policy.constants import (
    HASH_ALGORITHMS,
    KEYWORD_SOURCES,
    NONCE_PREFIX,
    NONE,
    SANDBOX_VALUES,
    SELF,
    WEBRTC_VALUES,
)
from content_security_policy.exceptions import BadDirectiveValue, BadSourceExpression
from content_security_policy.patterns import (
    BASE64_VALUE,
    HASH_SOURCE,
    HOST_SOURCE,
    # Appending _RE, so they are easier to distinguish from the constants
    KEYWORD_SOURCE as KEYWORD_SOURCE_RE,
    NONCE_SOURCE,
    NONE_SOURCE,
    NOT_SEPARATOR,
    SANDBOX_VALUE as SANDBOX_VALUE_RE,
    SCHEME,
    SCHEME_SOURCE,
    SELF_SOURCE,
    TOKEN,
    URI_REFERENCE,
    WEBRTC_VALUE as WEBRTC_VALUE_RE,
)
from content_security_policy.utils import AutoInstanceMixin


class SourceExpression(ValueItem, ABC):
    """
    Base class for all source expressions.
    """


# https://w3c.github.io/webappsec-csp/#grammardef-nonce-source
class NonceSrc(SourceExpression):
    pattern = NONCE_SOURCE

    def __init__(self, nonce: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        else:
            nonce = nonce.strip("'").lstrip(NONCE_PREFIX)
            if not BASE64_VALUE.fullmatch(nonce):
                raise BadSourceExpression(
                    f"Nonce value '{nonce}' does not match "
                    f"{BASE64_VALUE.pattern.pattern}"
                )
            value = f"'{NONCE_PREFIX}{nonce}'"

        super().__init__(value)


# https://w3c.github.io/webappsec-csp/#grammardef-nonce-source
class HashSrc(SourceExpression):
    pattern = HASH_SOURCE

    def __init__(
        self, hash_value: str, algo: Optional[str] = None, _value: Optional[str] = None
    ):
        if _value is not None:
            value = _value
        else:
            hash_value = hash_value.strip("'")
            if algo is not None:
                hash_value = hash_value
            else:
                algo, hash_value = hash_value.split("-")

            if algo not in HASH_ALGORITHMS:
                raise BadSourceExpression(f"Unknown hash algorithm: '{algo}'")

            if not BASE64_VALUE.fullmatch(hash_value):
                raise BadSourceExpression(
                    f"Hash value '{hash_value}' does not match "
                    f"{BASE64_VALUE.pattern.pattern}"
                )
            value = f"'{algo}-{hash_value}'"

        super().__init__(value)

    @classmethod
    def from_string(cls, str_value: str) -> HashSrc:
        return cls("", "", _value=str_value)


# https://w3c.github.io/webappsec-csp/#grammardef-scheme-source
class SchemeSrc(SourceExpression):
    pattern = SCHEME_SOURCE

    def __init__(self, scheme: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        else:
            scheme = scheme.rstrip(":")
            if not SCHEME.fullmatch(scheme):
                raise BadSourceExpression(
                    f"Scheme '{scheme}' does not match {SCHEME.pattern.pattern}"
                )
            value = f"{scheme}:"
        super().__init__(value)


# https://w3c.github.io/webappsec-csp/#grammardef-host-source
class HostSrc(SourceExpression):
    pattern = HOST_SOURCE

    def __init__(self, host: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        elif not self.pattern.fullmatch(host):
            raise BadSourceExpression(f"{host} does not match {self.pattern.pattern}")
        else:
            value = host
        super().__init__(value)


# https://w3c.github.io/webappsec-csp/#grammardef-keyword-source
class KeywordSource(AutoInstanceMixin, SourceExpression):
    pattern = KEYWORD_SOURCE_RE

    # You can later get an instance of any hash source by accessing these as class attributes
    # They are spelled out explicitly here so type hints work
    self = cast("KeywordSource", "'self'")
    unsafe_inline = cast("KeywordSource", "'unsafe-inline'")
    unsafe_eval = cast("KeywordSource", "'unsafe-eval'")
    strict_dynamic = cast("KeywordSource", "'strict-dynamic'")
    unsafe_hashes = cast("KeywordSource", "'unsafe-hashes'")
    report_sample = cast("KeywordSource", "'report-sample'")
    unsafe_allow_redirects = cast("KeywordSource", "'unsafe-allow-redirects'")
    wasm_unsafe_eval = cast("KeywordSource", "'wasm-unsafe-eval'")
    _auto_instance_prop = KEYWORD_SOURCES

    def __init__(self, keyword: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        else:
            no_ticks_keyword = keyword.strip("'")
            keyword = f"'{no_ticks_keyword}'"
            if not self.pattern.fullmatch(keyword):
                raise BadSourceExpression(
                    f"{keyword} does not match {self.pattern.pattern}"
                )
            value = keyword

        super().__init__(value)


class SingleValueItem(ClassAsValue, ValueItem, ABC):
    """
    Value item that only hase a single (semantic) value.
    Because they are case-insensitive, there still is a constructor for lenient parsing.
    """

    def __init__(self, *, _value: Optional[str] = None):
        value = _value or self._value
        super().__init__(value)

    @classmethod
    def from_string(cls: Type[SelfType], value: str) -> SelfType:
        return cls(_value=value)


# According to spec, 'none'  is not a `source-expression`, but a special case of `serialized-source-list`
# https://w3c.github.io/webappsec-csp/#grammardef-serialized-source-list
class NoneSrc(SingleValueItem):
    pattern = NONE_SOURCE
    _value = NONE


# Can be passed as class or an instance
NoneSrcType = Union[NoneSrc, Type[NoneSrc]]

# https://w3c.github.io/webappsec-csp/#grammardef-serialized-source-list
SourceList = Union[Tuple[SourceExpression], NoneSrcType]


class WebrtcValue(AutoInstanceMixin, ValueItem):
    # You can later get an instance of any hash by accessing these as class attributes
    # They are spelled out explicitly here so type hints work
    allow = cast("WebrtcValue", "'allow'")
    block = cast("WebrtcValue", "'block'")
    _auto_instance_prop = WEBRTC_VALUES

    def __init__(self, value: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        else:
            no_ticks_keyword = value.strip("'")
            value = f"'{no_ticks_keyword}'"
            if not WEBRTC_VALUE_RE.fullmatch(value):
                raise BadSourceExpression(
                    f"{value} does not match {KEYWORD_SOURCE_RE.pattern.pattern}"
                )

        super().__init__(_value)


# https://html.spec.whatwg.org/multipage/iframe-embed-object.html#the-iframe-elemet
class SandboxToken(AutoInstanceMixin, ValueItem):
    # You can later get an instance of any hash by accessing these as class attributes
    # They are spelled out explicitly here so type hints work
    allow_downloads = cast("SandboxToken", "allow-downloads")
    allow_forms = cast("SandboxToken", "allow-forms")
    allow_modals = cast("SandboxToken", "allow-modals")
    allow_orientation_lock = cast("SandboxToken", "allow-orientation-lock")
    allow_pointer_lock = cast("SandboxToken", "allow-pointer-lock")
    allow_popups = cast("SandboxToken", "allow-popups")
    allow_popups_to_escape_sandbox = cast(
        "SandboxToken", "allow-popups-to-escape-sandbox"
    )
    allow_presentation = cast("SandboxToken", "allow-presentation")
    allow_same_origin = cast("SandboxToken", "allow-same-origin")
    allow_scripts = cast("SandboxToken", "allow-scripts")
    allow_top_navigation = cast("SandboxToken", "allow-top-navigation")
    allow_top_navigation_by_user_activation = cast(
        "SandboxToken", "allow-top-navigation-by-user-activation"
    )
    allow_top_navigation_to_custom_protocols = cast(
        "SandboxToken", "allow-top-navigation-to-custom-protocols"
    )
    _auto_instance_prop = SANDBOX_VALUES

    def __init__(self, value: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        elif not SANDBOX_VALUE_RE.fullmatch(value):
            raise BadDirectiveValue(
                f"{value} does not match {SANDBOX_VALUE_RE.pattern.pattern}"
            )
        super().__init__(value)


# TODO: Unsure whether I like the Literal[""] here, need to revisit once I work on empty / non-empty directive-values
SandboxValue = Union[Tuple[SandboxToken], Literal[""]]


# 'self' is a keyword source expression, but it is also a possible hash for frame-ancestors, whereas other
# KeywordSources are not valid static_values for frame-ancestors.
class SelfSrc(SingleValueItem):
    pattern = SELF_SOURCE
    _value = SELF


# Can be passed as class or an instance
SelfSrcType = Union[SelfSrc, Type[SelfSrc]]

# https://w3c.github.io/webappsec-csp/#grammardef-ancestor-source-list
AncestorSource = Union[SchemeSrc, HostSrc, SelfSrcType]
AncestorSourceList = Union[Tuple[AncestorSource], NoneSrcType]


# https://w3c.github.io/webappsec-csp/#directive-report-to
class ReportToValue(ValueItem):
    pattern = TOKEN

    def __init__(self, value: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        elif not self.pattern.fullmatch(value):
            raise BadDirectiveValue(f"{value} does not match {self.pattern.pattern}")

        super().__init__(value)


# https://w3c.github.io/webappsec-csp/#directive-report-uri
class UriReference(ValueItem):
    pattern = URI_REFERENCE

    def __init__(self, value: str, _value: Optional[str] = None):
        if _value is not None:
            value = _value
        elif not self.pattern.fullmatch(value):
            raise BadDirectiveValue(f"{value} does not match {self.pattern.pattern}")

        super().__init__(value)


ReportUriValue = Tuple[UriReference]


class UnrecognizedValueItem(ValueItem):
    pattern = NOT_SEPARATOR



from content_security_policy.directives import *
from content_security_policy.values import *
from content_security_policy.base_classes import (
    Directive,
    ValueItem,
    ValueItemType,
    Policy,
    PolicyList,
)



__all__ = ["AutoCSPMiddleware"]

from typing import *

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http.response import HttpResponseBadRequest
from watchdog.observers import Observer

from content_security_policy import Directive, Policy, PolicyList
from content_security_policy.constants import (
    CSP_HEADER,
    CSP_RO_HEADER,
)
from content_security_policy.django.auto_src import AutoSrcDirective
from content_security_policy.django.constants import CSP_CONFIG_NAME, CSP_RO_CONFIG_NAME
from content_security_policy.django.utils.settings import get_csp_setting


class AutoCSPMiddleware:
    """
    Uses the CONTENT_SECURITY_POLICY setting to build a Content-Security-Policy (CSP)
    for the site. Sets up watchers to update dynamic parts of the CSP. Injects CSP
    header into every response.
    """

    def __init__(self, get_response):
        if not settings.DEBUG:
            raise ImproperlyConfigured(
                f"{self.__class__.__name__} may only be used with DEBUG = True."
            )

        self.get_response = get_response
        self.policy_lists: Dict[
            str, Tuple[Tuple[Directive | AutoSrcDirective, ...], ...]
        ] = {}

        for config_name, header_name in (
            (CSP_CONFIG_NAME, CSP_HEADER),
            (CSP_RO_CONFIG_NAME, CSP_RO_HEADER),
        ):
            if getattr(settings, config_name, None):
                self.policy_lists[header_name] = get_csp_setting(config_name)

        if not self.policy_lists:
            raise ImproperlyConfigured(
                f"{self.__class__.__name__} used but neither {CSP_CONFIG_NAME} nor "
                f"{CSP_RO_CONFIG_NAME} is set in settings."
            )

        for policy_list in self.policy_lists.values():
            for policy in policy_list:
                for directive in policy:
                    if isinstance(directive, AutoSrcDirective):
                        try:
                            observer = self.observer
                        except AttributeError:
                            observer = self.observer = Observer()
                        for path in directive.watch_dirs:
                            observer.schedule(directive, path, recursive=True)

        if hasattr(self, "observer"):
            self.observer.start()

    @staticmethod
    def render(policies, scheme: str, host: str):
        return PolicyList(
            *(
                Policy(
                    *(
                        directive.render(scheme=scheme, host=host)
                        if isinstance(directive, AutoSrcDirective)
                        else directive
                        for directive in directives
                    )
                )
                for directives in policies
            )
        )

    def __call__(self, request):
        response = self.get_response(request)
        scheme = request.scheme
        try:
            host = request.META["HTTP_HOST"]
        except KeyError:
            return HttpResponseBadRequest("Host Header Missing from request.")

        for header, policies in self.policy_lists.items():
            response[header] = self.render(policies, scheme, host)

        return response



"""
Actual directives, I would have loved to generate these classes dynamically, but then autocompletion tools won't
properly pick up on them.
"""
__all__ = [
    "SourceListDirective",
    "ChildSrc",
    "ConnectSrc",
    "DefaultSrc",
    "FontSrc",
    "FrameSrc",
    "ImgSrc",
    "ManifestSrc",
    "MediaSrc",
    "ObjectSrc",
    "ScriptSrc",
    "ScriptSrcElem",
    "ScriptSrcAttr",
    "StyleSrc",
    "StyleSrcElem",
    "StyleSrcAttr",
    "BaseUri",
    "Sandbox",
    "FormAction",
    "FrameAncestors",
    "ReportUri",
    "ReportTo",
    "Webrtc",
    "WorkerSrc",
    "UnrecognizedDirective",
]

from abc import ABC
from typing import Optional, Union

from content_security_policy.base_classes import (Directive, SelfType, SingleValueDirective)
from content_security_policy.exceptions import (
    BadDirectiveValue,
    BadSourceList,
)
from content_security_policy.values import (AncestorSource, AncestorSourceList, NoneSrc, NoneSrcType, ReportToValue,
                                            ReportUriValue, SandboxValue, SourceExpression, SourceList,
                                            UnrecognizedValueItem)


# This is not called FetchDirective because not all directives accepting a Source List are categorised as
# Fetch Directives by the spec (worker-src, base-uri, form-action)
class SourceListDirective(Directive[SourceList], ABC):
    """
    A directive whose hash is a
    """

    def __init__(self, *sources: Union[SourceExpression, NoneSrcType], **kwargs):
        if len(sources) > 1 and any(src == NoneSrc for src in sources):
            raise BadSourceList(
                f"{NoneSrc} may not be combined with other source expressions."
            )
        super().__init__(*sources, **kwargs)

    def __add__(self: SelfType, other: SourceExpression) -> SelfType:
        return type(self)(*self.values, other, name=self._name)


# Fetch Directives
class ChildSrc(SourceListDirective):
    _name = "child-src"


class ConnectSrc(SourceListDirective):
    _name = "connect-src"


class DefaultSrc(SourceListDirective):
    _name = "default-src"


class FontSrc(SourceListDirective):
    _name = "font-src"


class FrameSrc(SourceListDirective):
    _name = "frame-src"


class ImgSrc(SourceListDirective):
    _name = "img-src"


class ManifestSrc(SourceListDirective):
    _name = "manifest-src"


class MediaSrc(SourceListDirective):
    _name = "media-src"


class ObjectSrc(SourceListDirective):
    _name = "object-src"


class ScriptSrc(SourceListDirective):
    _name = "script-src"


class ScriptSrcElem(SourceListDirective):
    _name = "script-src-elem"


class ScriptSrcAttr(SourceListDirective):
    _name = "script-src-attr"


class StyleSrc(SourceListDirective):
    _name = "style-src"


class StyleSrcElem(SourceListDirective):
    _name = "style-src-elem"


class StyleSrcAttr(SourceListDirective):
    _name = "style-src-attr"


# Other directives
class Webrtc(SingleValueDirective[SourceList]):
    _name = "webrtc"


class WorkerSrc(SourceListDirective):
    _name = "worker-src"


# Document directives
class BaseUri(SourceListDirective):
    _name = "base-uri"


class Sandbox(Directive[SandboxValue]):
    _name = "sandbox"


# Navigation directives
class FormAction(SourceListDirective):
    _name = "form-action"


class FrameAncestors(Directive[AncestorSourceList]):
    _name = "frame-ancestors"

    def __init__(self, *sources: Union[AncestorSource, NoneSrcType], **kwargs):
        """
        Create frame-ancestors from NoneSrc XOR an arbitrary number of AncestorSource.
        :param sources: Allowed frame ancestors.
        """
        if len(sources) > 1 and any(src == NoneSrc for src in sources):
            raise BadDirectiveValue(
                f"{NoneSrc} may not be combined with other ancestor sources."
            )
        super().__init__(*sources, **kwargs)


# Reporting directives
class ReportUri(Directive[ReportUriValue]):
    _name = "report-uri"


class ReportTo(SingleValueDirective[ReportToValue]):
    _name = "report-to"


class UnrecognizedDirective(Directive[UnrecognizedValueItem]):
    """
    A directive whose name is not recognized.
    """

    def __init__(self, *values, name: Optional[str] = None, **kwargs):
        if name is None and "_name" not in kwargs:
            raise ValueError(
                f"You must pass either '_name' or 'name' as a kwarg to "
                f"{type(self).__name__}. 'name' takes precedence."
            )

        if name is not None:
            kwargs["_name"] = name
        super().__init__(*values, **kwargs)

    @property
    def name(self) -> str:
        return self._name



"""
Django settings for content_security_policy.django project.
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-t-3nx1^9ves=v(0+2sg)#lo3u^6m&)7pydi-vnf(8z)2s7x0lg"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "content_security_policy.django",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "content_security_policy.django.middleware.AutoCSPMiddleware",
]

ROOT_URLCONF = "content_security_policy.django.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "content_security_policy.django.wsgi.application"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.2/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


from content_security_policy.directives import *
from content_security_policy.values import *
from content_security_policy.django.auto_src import AutoHostScriptSrc


EXTERNAL_SCRIPTS = [HostSrc("https://code.jquery.com/jquery-3.5.1.js")]

CONTENT_SECURITY_POLICY = [
    AutoHostScriptSrc(*EXTERNAL_SCRIPTS, watch_apps=INSTALLED_APPS),
    ScriptSrcAttr(NoneSrc),
    DefaultSrc(KeywordSource.self),
    StyleSrc(KeywordSource.self),
    FrameAncestors(NoneSrc),
    BaseUri(NoneSrc),
    ObjectSrc(NoneSrc),
    ManifestSrc(NoneSrc),
]



__all__ = ["value_item_from_string", "directive_from_string", "policy_from_string"]

from typing import *
from content_security_policy import *
from content_security_policy.exceptions import ParsingError, NoSuchDirective
from content_security_policy.patterns import (
    VALUE_ITEM_SEPARATOR,
    DIRECTIVE_SEPARATOR,
    POLICY_SEPARATOR,
    WHITESPACE_HEAD,
)

_PARSING_RULES: Dict[Type[Directive], Tuple[Type[ValueItem]]] = {
    UnrecognizedDirective: tuple(),
    SourceListDirective: (
        NoneSrc,
        KeywordSource,
        HashSrc,
        NonceSrc,
        HostSrc,
        SchemeSrc,
    ),
    Webrtc: (WebrtcValue,),
    Sandbox: (SandboxToken,),
    FrameAncestors: (NoneSrc, SelfSrc, HostSrc, SchemeSrc),
    ReportUri: (UriReference,),
    ReportTo: (ReportToValue,),
}


def value_item_from_string(
    value_string: str, directive_type: Type[Directive]
) -> ValueItemType:
    """
    Create a directive hash object from a string.
    :param value_string: Directive hash (without whitespace!)
    :param directive_type: Directive which has hash, needed to distinguish certain hash types.
    :return: Object representing the hash.
    """
    for d_type, value_types in _PARSING_RULES.items():
        try:
            if issubclass(directive_type, d_type):
                for v_type in value_types:
                    if v_type.pattern.fullmatch(value_string):
                        v_type = cast(Type[ValueItem], v_type)
                        return v_type.from_string(value_string)

                return UnrecognizedValueItem(value_string)
        except Exception as e:
            print(e)

    raise ValueError(
        f"Failed to find parsing rules for directive type {directive_type}"
    )


def directive_from_string(directive_string: str) -> Directive:
    separators = VALUE_ITEM_SEPARATOR.findall(directive_string)
    tokens = VALUE_ITEM_SEPARATOR.split(directive_string)
    if len(separators) != (len(tokens) - 1):
        raise ParsingError(
            "Mismatch in amount of tokens and separators. "
            "Perhaps your directive is not trimmed?"
        )
    if len(tokens) == 0:
        raise ParsingError("No directive name found in directive string.")

    name, value_items = tokens[0], tokens[1:]
    try:
        dir_class = Directive.class_by_name(name.lower())
    except NoSuchDirective:
        dir_class = UnrecognizedDirective
    dir_class = cast(Type[Directive], dir_class)

    values = (
        value_item_from_string(item, directive_type=dir_class) for item in value_items
    )
    return dir_class(*values, _name=name, _separators=tuple(separators))


def policy_from_string(policy_string: str) -> Policy:
    separators = DIRECTIVE_SEPARATOR.findall(policy_string)
    tokens = DIRECTIVE_SEPARATOR.split(policy_string)
    if len(separators) != (len(tokens) - 1):
        raise ParsingError(
            "Mismatch in amount of tokens and separators. "
            "Perhaps your policy is not trimmed?"
        )

    return Policy(
        *(directive_from_string(dir) for dir in tokens), _separators=separators
    )


def policy_list_from_string(policy_list_string: str) -> PolicyList:
    head = WHITESPACE_HEAD.match(policy_list_string).group(0)
    policy_list_string = WHITESPACE_HEAD.sub("", policy_list_string)

    # I have given up on a regex replace for trailing whitespace...
    policy_list_string = policy_list_string[::-1]
    tail = WHITESPACE_HEAD.match(policy_list_string).group(0)[::-1]
    policy_list_string = WHITESPACE_HEAD.sub("", policy_list_string)
    policy_list_string = policy_list_string[::-1]

    separators = POLICY_SEPARATOR.findall(policy_list_string)
    tokens = POLICY_SEPARATOR.split(policy_list_string)
    return PolicyList(
        *(policy_from_string(pol) for pol in tokens),
        _head=head,
        _tail=tail,
        _separators=separators,
    )



from unittest import TestCase

from content_security_policy import directives
from content_security_policy.utils import kebab_to_pascal
from content_security_policy.constants import DIRECTIVE_NAMES


class DirectivesComplete(TestCase):
    def test_directives_complete(self):
        for name in DIRECTIVE_NAMES:
            with self.subTest(name):
                class_name = kebab_to_pascal(name)
                if not hasattr(directives, class_name):
                    self.fail(
                        f"directives has no class {class_name} for directive {name}"
                    )



from __future__ import annotations

from abc import ABC
from functools import cache, cached_property
import re
from typing import Any, Dict, Generic, Iterable, Optional, Tuple, Type, TypeVar, Union

from content_security_policy.constants import (
    DEFAULT_DIRECTIVE_SEPARATOR,
    DEFAULT_POLICY_SEPARATOR,
    DEFAULT_VALUE_SEPARATOR,
)
from content_security_policy.exceptions import NoSuchDirective
from content_security_policy.utils import StrOnClassMeta, kebab_to_snake


class ValueItem(ABC):
    """
    Base class for "items" in directive values. To clarify the distinction from a directive hash, consider:
    script-src 'self' http://example.com
    The hash of this script-src directive is "'self' http://example.com", whereas "'self'" and "http://example.com"
    are the "items" of the hash.
    The two are not mutually exclusive. Some items may also be valid values by themselves.
    """

    # Pattern used to identify these values when parsing
    pattern: re.Pattern
    # hash as string
    _value: Optional[str]

    def __init__(self, value: str, _value: Optional[str] = None):
        # The arguments are this awkward because all concrete implementations of ValueItem use their first argument
        # for the "human friendly" construction of values. The kw arg "_value" is then used to create values from
        # strings. Both are included here so the type-checker recognizes that all ValueItem constructors accept the
        # _value kw_arg.
        self._value = value or _value

    def __str__(self):
        return self._value

    @classmethod
    def from_string(cls: Type[SelfType], str_value: str) -> SelfType:
        """
        Return an instance of the class from a string value using the `_value` kwarg.
        Note that __init__ on subclasses of ValueItem MUST Ignore any other arguments if _value is passed.
        ValueItem subclasses MUST override this method if their __init__ has a different argument structure.
        :param str_value: value to pass to _value.
        :return: Instance of the class.
        """
        return cls("", _value=str_value)


class ClassAsValue(ValueItem, metaclass=StrOnClassMeta):
    def __str__(self):
        """
        Calling str() on an instance will return the hash.
        """
        return self._value


# Some classes for directive values are valid items themselves, use this to cover both in type hints
ValueItemType = Union[ValueItem, Type[ValueItem]]

SelfType = TypeVar("SelfType", bound="Directive")
ValueType = TypeVar("ValueType", bound=ValueItemType)

_DIRECTIVE_REGISTER: Dict[str, Type[Directive]] = {}


class Directive(ABC, Generic[ValueType]):
    _value: Tuple[ValueType]
    _name: Optional[str]
    _separators: Tuple[str]

    def __init__(
        self,
        *values: ValueType,
        _name: Optional[str] = None,
        _separators: Optional[Iterable[str]] = None,
    ):
        self._value = tuple(values)

        self._separators = (
            tuple(_separators)
            if _separators is not None
            else ((DEFAULT_VALUE_SEPARATOR,) * len(self._value))
        )

        if _name is not None:
            self._name = _name

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        names = {cls.__name__}
        if hasattr(cls, "_name"):
            names.add(cls._name)
            names.add(kebab_to_snake(cls._name))
        for name in names:
            if name in _DIRECTIVE_REGISTER:
                raise ValueError(f"{name} already registered in directive register.")
            _DIRECTIVE_REGISTER[name] = cls

    @classmethod
    def class_by_name(cls, directive_name: str) -> Type[Directive]:
        for name in (directive_name, directive_name.lower()):
            if name in _DIRECTIVE_REGISTER:
                return _DIRECTIVE_REGISTER[name]

        raise NoSuchDirective(
            f"Can not find class for directive _name {directive_name}"
        )

    @property
    def name(self) -> str:
        """
        Name of the directive.
        """
        return self._name

    @cached_property
    def values(self) -> Tuple[ValueType]:
        """
        All values of the directive as a tuple
        :return:
        """
        return self._value if isinstance(self._value, tuple) else (self._value,)

    @property
    def _value_str_tokens(self):
        value_it = iter(self.values)
        try:
            yield str(next(value_it))
        except StopIteration:
            # Happens if self.values is empty (no-value directives)
            return
        for sep, value in zip(self._separators[1:], value_it):
            yield sep
            yield str(value)

    @property
    def _str_tokens(self):
        yield self.name
        if self._separators:
            yield self._separators[0]
        yield from self._value_str_tokens

    @cached_property
    def value(self) -> str:
        """
        Return the complete value of the directive as a string
        :return:
        """
        return "".join(self._value_str_tokens)

    @cache
    def __str__(self):
        return "".join(self._str_tokens)

    def __iter__(self):
        """
        Iterate the directives values.
        """
        yield from self.values

    def __add__(self: SelfType, other: ValueType) -> SelfType:
        separators = self._separators + (DEFAULT_VALUE_SEPARATOR,)
        return type(self)(*self.values, other, _separators=separators, _name=self._name)

    def __sub__(self: SelfType, other: ValueType) -> SelfType:
        raise NotImplemented


class SingleValueDirective(Directive[ValueType], ABC, Generic[ValueType]):
    """
    A directive that only supports one value item.
    """

    # __init__ still allows for multiple values to be passed to enable lenient parsing.
    # TODO: When implementing is_valid, do not forget to check whether there is more than one value!
    def __init__(self, *values, **kwargs):
        super().__init__(*values, **kwargs)


class Policy:
    _separators: Optional[Tuple[str]]

    def __init__(
        self,
        *directives,
        _separators: Optional[Iterable[str]] = None,
    ):
        self._directives = tuple(directives)
        self._separators = (
            tuple(_separators)
            if _separators is not None
            else ((DEFAULT_DIRECTIVE_SEPARATOR,) * (len(self._directives) - 1))
        )

    @property
    def directives(self):
        return self._directives

    @property
    def _str_tokens(self):
        directives_it = iter(self.directives)
        yield str(next(directives_it))
        for sep, directive in zip(self._separators, directives_it):
            yield sep
            yield str(directive)

    @cache
    def __str__(self):
        return "".join(self._str_tokens)

    @cache
    def _get_indices(self, directive_type: Type[Directive]) -> Tuple[int]:
        indices = []
        for i, directive in enumerate(self.directives):
            if isinstance(directive, directive_type):
                indices.append(i)

        return tuple(indices)

    @cache
    def __getitem__(self, key: Union[Type[Directive], int, str]) -> Directive:
        """
        Get a directive of the policy. If key is an int, the key-th directive in the policy is returned.
        If key is a string or a Directive type, the FIRST directive of that type is returned.
        String keys are case-insensitive and support both PascalCase and kebab-case (e.g. ScriptSrc and script-src)
        :param key: selector for directive.
        :return: The selected directive.
        """
        if type(key) is int:
            return self.directives[key]

        if isinstance(key, str):
            key = Directive.class_by_name(key)
        elif not issubclass(key, Directive):
            raise TypeError(
                f"Item must be either an int, a string or a subclass of {Directive.__name__}, not {type(key)}"
            )

        for directive in self.directives:
            if isinstance(directive, key):
                return directive

        raise IndexError(f"Policy does not have a {key.__name__} directive.")

    def __getattr__(self, name: str) -> Any:
        """
        Get a directive of the policy.
        :param name: Directive name like ScriptSrc or script-src
        :return: First instance of the specified directive type in this policy.
        """
        if name in _DIRECTIVE_REGISTER:
            try:
                return self.__getitem__(name)
            except IndexError as e:
                raise AttributeError(*e.args)
        else:
            return object.__getattribute__(self, name)

    def __iter__(self):
        """
        Iterate over directives.
        """
        yield from self.directives

    def __add__(self, other: Directive) -> Policy:
        """
        Add a new directive to the policy.
        :param other:
        :return:
        """
        separators = self._separators + (DEFAULT_DIRECTIVE_SEPARATOR,)
        return type(self)(*self.directives, other, _separators=separators)

    def __sub__(self, other: Type[Directive]) -> Policy:
        """
        Get a copy of the policy with all directives of the specified type removed.
        :param other: Type of directive to remove.
        :return: New policy with no directive of type other.
        """
        new_directives = []
        new_separators = []
        removed_indices = self._get_indices(other)
        for i, directive in enumerate(self.directives):
            if i not in removed_indices:
                new_directives.append(directive)
                if i != 0 and len(new_directives):
                    new_separators.append(self._separators[i - 1])

        return type(self)(*new_directives, _separators=new_separators)

    def __and__(self, other: Policy) -> PolicyList:
        return PolicyList(self, other)


class PolicyList:
    _policies: Tuple[Policy]
    _separators: Tuple[str]

    def __init__(
        self,
        *policies: Policy,
        _separators: Optional[Iterable[str]] = None,
        _head: Optional[str] = None,
        _tail: Optional[str] = None,
    ):
        self._policies = tuple(policies)
        self._separators = (
            tuple(_separators)
            if _separators is not None
            else ((DEFAULT_POLICY_SEPARATOR,) * (len(self._policies) - 1))
        )
        self._head = _head
        self._tail = _tail

    @property
    def _str_tokens(self):
        if self._head:
            yield self._head

        policy_it = iter(self._policies)
        yield str(next(policy_it))

        for sep, policy in zip(self._separators, policy_it):
            yield sep
            yield str(policy)

        if self._tail:
            yield self._tail

    @cache
    def __str__(self):
        return "".join(self._str_tokens)



from argparse import ArgumentParser
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, List, Tuple

from django.conf import settings
from django.core.checks import Tags
from django.core.exceptions import ImproperlyConfigured
from django.core.files import File
from django.core.management.base import BaseCommand, CommandError

from content_security_policy import Policy, PolicyList
from content_security_policy.django.auto_src import AutoSrcDirective
from content_security_policy.django.constants import CSP_CONFIG_NAME, CSP_RO_CONFIG_NAME
from content_security_policy.django.exceptions import ValuesMissing
from content_security_policy.django.utils.settings import get_csp_setting

_CSP = "csp"
_CSP_RO = "csp_ro"

_NAME_OPT = "--name"
_NAME_RO_OPT = "--name_ro"


class Command(BaseCommand):
    help = (
        f"Build CSP based on {CSP_CONFIG_NAME} and {CSP_RO_CONFIG_NAME} settings "
        "and save them to plain text files."
    )
    requires_system_checks = [Tags.staticfiles]

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            _NAME_OPT,
            default="csp",
            dest=f"{_CSP}_name",
            help="Staticfiles file name to use for CSP file.",
        )
        parser.add_argument(
            _NAME_RO_OPT,
            default="csp_ro",
            dest=f"{_CSP_RO}_name",
            help="Staticfiles file name to use for report-only CSP file.",
        )
        parser.add_argument(
            "--path",
            help="Write CSP to path instead of staticfiles storage. "
            f"{_NAME_OPT} will be ignored.",
            type=Path,
            default=None,
            dest=f"{_CSP}_path",
        )
        parser.add_argument(
            "--path_ro",
            help="Write report-only CSP to path instead of staticfiles storage. "
            f"{_NAME_RO_OPT} will be ignored.",
            type=Path,
            default=None,
            dest=f"{_CSP_RO}_path",
        )

    def render(self) -> Dict[str, PolicyList]:
        """
        Render csp settings into csp classes.
        Produce useful error messages if applicable
        """
        policy_lists = {
            name: get_csp_setting(config_name)
            for name, config_name in (
                (_CSP, CSP_CONFIG_NAME),
                (_CSP_RO, CSP_RO_CONFIG_NAME),
            )
            if getattr(settings, config_name, None)
        }

        if not policy_lists:
            raise ImproperlyConfigured(
                f"{self.__class__.__name__} called but neither {CSP_CONFIG_NAME} nor "
                f"{CSP_RO_CONFIG_NAME} found in settings."
            )

        errors: List[Tuple[AutoSrcDirective, ValuesMissing]] = []
        rendered_policy_lists = {}
        for policy_type, policies in policy_lists.items():
            rendered_policies = []
            for policy in policies:
                rendered_directives = []
                for directive in policy:
                    if isinstance(directive, AutoSrcDirective):
                        # Auto directives do not get any additional args when they are
                        # rendered through the buildcsp command.
                        # This is intentional, you are forced to have everything in
                        # one place.
                        try:
                            rendered_directives.append(directive.render())
                        except ValuesMissing as e:
                            errors.append((directive, e))
                    else:
                        rendered_directives.append(directive)
                rendered_policies.append(Policy(*rendered_directives))
            rendered_policy_lists[policy_type] = PolicyList(*rendered_policies)

        if errors:
            message = "Your CSP can not be rendered.\n"
            for directive, e in errors:
                miss_names = ", ".join(e.names)
                message += (
                    f"Directive {directive.name} is missing values "
                    f"for: {miss_names}.\n"
                )

            raise CommandError(message, returncode=-1)

        return rendered_policy_lists

    def handle(self, *args, **options):
        rendered_policy_lists = self.render()

        for name, policy_list in rendered_policy_lists.items():
            file_name = options[f"{name}_name"]
            path: Path = options[f"{name}_path"]

            if path:
                with open(path, "w") as f:
                    f.write(str(policy_list))
                self.stdout.write(f"Wrote {name} value to {path.absolute()}")
            else:
                # This import is here because otherwise you would need to configure
                # staticfiles even if you don't use it.
                from django.contrib.staticfiles.storage import staticfiles_storage

                with NamedTemporaryFile("w+") as tmp:
                    tmp.write(str(policy_list))
                    tmp.flush()
                    with File(tmp) as django_f:
                        if staticfiles_storage.exists(file_name):
                            self.stdout.write(
                                f"Deleting file {staticfiles_storage.path(file_name)}"
                            )
                            staticfiles_storage.delete(file_name)

                        staticfiles_storage.save(file_name, django_f)
                        self.stdout.write(
                            f"Saved CSP at {staticfiles_storage.path(file_name)}"
                        )



from unittest import TestCase

from content_security_policy import (
    NoneSrc,
    KeywordSource,
    SelfSrc,
    FrameAncestors,
    Directive,
)
from content_security_policy.constants import SOURCE_LIST_DIRECTIVES
from content_security_policy.exceptions import BadSourceList, BadDirectiveValue


class DoNotCombineNoneSource(TestCase):
    def test_source_list_directives_constructor(self):
        for directive in SOURCE_LIST_DIRECTIVES:
            with self.subTest(directive):
                directive_cls = Directive.class_by_name(directive)
                with self.assertRaises(BadSourceList):
                    directive_cls(NoneSrc, KeywordSource.self)

    def test_source_list_directives_addition(self):
        for directive in SOURCE_LIST_DIRECTIVES:
            with self.subTest(directive):
                directive_cls = Directive.class_by_name(directive)
                instance = directive_cls(NoneSrc)
                with self.assertRaises(BadSourceList):
                    instance += KeywordSource.self

    def test_frame_ancestors_constructor(self):
        with self.assertRaises(BadDirectiveValue):
            FrameAncestors(NoneSrc, SelfSrc)
